#include <DBManager.h>

#include <ctime>

#include <boost/date_time.hpp>
#include <boost/lexical_cast.hpp>

#include <soci/mysql/soci-mysql.h>

#include <mysql.h>

namespace MetadataImporter_ns
{

//==============================================================================
//      DBManager::DBManager()
//==============================================================================
DBManager::DBManager(Tango::DeviceImpl* deviceImpl_p,
    Configuration::SP configuration_sp) : Tango::LogAdapter(deviceImpl_p),
    m_configuration_sp(configuration_sp)
{
    DEBUG_STREAM << "DBManager::DBManager()" << endl;

    m_session_sp.reset(new soci::session);
}

//==============================================================================
//      DBManager::DBManager()
//==============================================================================
DBManager::~DBManager()
{
    DEBUG_STREAM << "DBManager::~DBManager()" << endl;

    m_session_sp->close();
}

//==============================================================================
//      DBManager::DBManager()
//==============================================================================
DBManager::SP DBManager::create(Tango::DeviceImpl* deviceImpl_p,
    Configuration::SP configuration_sp)
{
    DBManager::SP d_sp(new DBManager(deviceImpl_p, configuration_sp),
        DBManager::Deleter());

    return d_sp;
}

//==============================================================================
//      DBManager::connect()
//==============================================================================
void DBManager::connect() throw(soci::soci_error)
{
    DEBUG_STREAM << "DBManager::connect()" << endl;

    boost::mutex::scoped_lock lock(m_connectionMutex);

    std::stringstream connection;
    connection << " host=" << m_configuration_sp->getDatabaseHost();
    connection << " port=" << m_configuration_sp->getDatabasePort();
    connection << " user=" << m_configuration_sp->getDatabaseUsername();
    connection << " password=" << m_configuration_sp->getDatabasePassword();

    #ifdef VERBOSE_DEBUG
    INFO_STREAM << "CONNECTION: " << connection.str() << endl;
    #endif

    m_session_sp->open(soci::mysql, connection.str());

    soci::mysql_session_backend* backend =
            static_cast<soci::mysql_session_backend*>(m_session_sp->get_backend());

    MYSQL* mysql = backend->conn_;

    bool reconnect = true;
    mysql_options(mysql, MYSQL_OPT_RECONNECT, &reconnect);
}

//==============================================================================
//      DBManager::disconnect()
//==============================================================================
void DBManager::disconnect()
{
    DEBUG_STREAM << "DBManager::disconnect()" << endl;

    boost::mutex::scoped_lock lock(m_connectionMutex);

    m_session_sp->close();
}

//==============================================================================
//      DBManager::retrieveInformation()
//==============================================================================
DBManager::InformationList DBManager::retrieveInformation(std::string schema,
    std::string table) throw(soci::soci_error)
{
    DEBUG_STREAM << "DBManager::retrieveInformation()" << endl;

    boost::mutex::scoped_lock lock(m_connectionMutex);

    if(m_session_sp->get_backend() == NULL)
        m_session_sp->reconnect();

    soci::rowset<InformationTuple> rows = (m_session_sp->prepare << "select "
        "column_name, column_type, is_nullable from information_schema.columns "
        "where table_schema like :schema and table_name like :table",
        soci::use(schema, "schema"), soci::use(table, "table") );

    InformationList informationList;

    std::copy(rows.begin(), rows.end(), std::back_inserter(informationList));

    if(informationList.empty())
    {
        std::stringstream errorStream;
        errorStream << schema << "." << table << " not exists";
        throw soci::soci_error(errorStream.str());
    }

    return informationList;
}

//==============================================================================
//      DBManager::retrieveLastTimestamp()
//==============================================================================
std::tm DBManager::retrieveLastTimestamp(std::string schema, std::string table)
        throw(soci::soci_error)
{
    DEBUG_STREAM << "DBManager::retrieveLastTimestamp()" << endl;

    boost::mutex::scoped_lock lock(m_connectionMutex);

    if(m_session_sp->get_backend() == NULL)
        m_session_sp->reconnect();

    std::tm lastTimestamp;

    *m_session_sp << "select coalesce(max(update_time),'1970-01-01 00:00:00') "
        "from "<< schema << "." << table, soci::into(lastTimestamp);

    return lastTimestamp;
}

//==============================================================================
//      DBManager::persistMetadata()
//==============================================================================
void DBManager::persistMetadata(std::string schema, std::string table,
    const Response::Metadata& metadata) throw(soci::soci_error, std::out_of_range)
{
    DEBUG_STREAM << "DBManager::persistMetadata()" << endl;

    boost::mutex::scoped_lock lock(m_connectionMutex);

    if(m_session_sp->get_backend() == NULL)
        m_session_sp->reconnect();

    soci::transaction transaction(*m_session_sp);

    for(int i=0; i<metadata.rows_size(); ++i)
        *m_session_sp << composeInsertQuery(schema, table, metadata.rows(i));

    transaction.commit();
}

//==============================================================================
//      DBManager::composeInsertQuery()
//==============================================================================
std::string DBManager::composeInsertQuery(std::string schema, std::string table,
    const Response::Metadata::Row& row) throw(std::out_of_range)
{
    DEBUG_STREAM << "DBManager::composeInsertQuery()" << endl;


	//TODO: check if all lexical cast are necessary

    std::stringstream keysStream;
    std::stringstream valuesStream;
    std::stringstream keyValuesStream;

    std::string databaseTable = m_configuration_sp->getDatabaseTable();

    //DtDate list
    for(int i = 0; i<row.date_list_size(); ++i)
    {
        int64_t rawTimestamp = row.date_list(i).value();
        std::tm tmTimestamp = *localtime(&rawTimestamp);
        boost::posix_time::ptime timestamp =
                boost::posix_time::ptime_from_tm(tmTimestamp);

        keysStream << databaseTable << "." << row.date_list(i).key() << ",";
        valuesStream << "'" << boost::posix_time::to_iso_string(timestamp) << "',";
        keyValuesStream << databaseTable << "." << row.date_list(i).key()
            << "='" << boost::posix_time::to_iso_string(timestamp) << "',";
    }

    //DtDouble list
    for(int i = 0; i<row.double_list_size(); ++i)
    {
        keysStream << databaseTable << "." << row.double_list(i).key() << ",";
        valuesStream << boost::lexical_cast<std::string>(row.double_list(i).value()) << ",";
        keyValuesStream << databaseTable << "." << row.double_list(i).key()
            << "=" << row.double_list(i).value() << ",";
    }

    //DtInteger list
    for(int i = 0; i<row.integer_list_size(); ++i)
    {
        keysStream << databaseTable << "." << row.integer_list(i).key() << ",";
        valuesStream << boost::lexical_cast<std::string>(row.integer_list(i).value()) << ",";
        keyValuesStream << databaseTable << "." << row.integer_list(i).key()
            << "=" << row.integer_list(i).value() << ",";
    }

    //DtLongLong list
    for(int i = 0; i<row.long_long_list_size(); ++i)
    {
        keysStream << databaseTable << "." << row.long_long_list(i).key() << ",";
        valuesStream << boost::lexical_cast<std::string>(row.long_long_list(i).value()) << ",";
        keyValuesStream << databaseTable << "." << row.long_long_list(i).key()
            << "=" << row.long_long_list(i).value() << ",";
    }

    //DtString list
    for(int i = 0; i<row.strings_list_size(); ++i)
    {
        keysStream << databaseTable << "." << row.strings_list(i).key() << ",";
        valuesStream << "'" << row.strings_list(i).value() << "',";
        keyValuesStream << databaseTable << "." << row.strings_list(i).key()
            << "='" << row.strings_list(i).value() << "',";
    }

    //DtUnsignedLong list
    for(int i = 0; i<row.unsinged_long_list_size(); ++i)
    {
        keysStream << databaseTable << "." << row.unsinged_long_list(i).key() << ",";
        valuesStream << boost::lexical_cast<std::string>(row.unsinged_long_list(i).value()) << ",";
        keyValuesStream << databaseTable << "." << row.unsinged_long_list(i).key()
            << "=" << row.unsinged_long_list(i).value() << ",";
    }

    std::string keys = keysStream.str();
    if(keys.at(keys.size()-1)==',')
        keys.erase(keys.size()-1);

    std::string values = valuesStream.str();
    if(values.at(values.size()-1)==',')
        values.erase(values.size()-1);

    std::string keyValues = keyValuesStream.str();
    if(keyValues.at(keyValues.size()-1)==',')
        keyValues.erase(keyValues.size()-1);

    /*
     * 2015/03/17 error: maybe an erroneus return value from mysql in function (retrieveLastTimestamp)
     * triggered an update of all metadata tables simultaneously. Still under investigation.
     * Skip storage_path and file_path to null will not wait file to be downloaded after a metadata update.
     * This is temporary workaraound and exclude admin console to work properly.
     *
    stringstream query;
    query << "INSERT INTO " << schema << "." << table
        << " (" << keys << ") VALUES (" << values << ") "
        << " ON DUPLICATE KEY UPDATE storage_path=NULL, file_path=NULL, " << keyValues;
    */

    stringstream query;
    query << "INSERT INTO " << schema << "." << table
    << " (" << keys << ") VALUES (" << values << ") "
    << " ON DUPLICATE KEY UPDATE " << keyValues;

    #ifdef VERBOSE_DEBUG
        INFO_STREAM << "DBManager::composeInsertQuery() " << query.str() << endl;
    #endif

    return query.str();
}

}   //namespace
